/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Jack Lo.
Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
 */

#include "main.h"
#include "predict.h"
#include "smt.h"
#include "mem.h"
#include "cache.h"
#include "inst.h"

scoreboard_t *exec_head;
scoreboard_t *exec_tail;
int num_in_exec;

extern scoreboard_t *iq_head;
extern scoreboard_t *iq_tail;
extern scoreboard_t *fq_head;
extern scoreboard_t *fq_tail;
extern int simulating;

extern scoreboard_t *rwrite_head;

extern int num_in_rwrite;

extern int numiniq, numinfq;

extern int num_in_decode, num_in_rename, num_in_rread;
extern scoreboard_t *decode_head, *decode_tail, *rename_head,
  *rread_head, *rread_tail;
extern cache_request *cache_queue_head;
extern int debugsync;
extern int numcontexts;
extern unsigned long cyc;
extern context *thecontexts[];
extern int priority[];
extern int key[];

extern int i_registers_freed;
extern int f_registers_freed;

acq_entry *acquire_queue[MAXTHREADS];
static int lock_flag[MAXTHREADS];
static unsigned long lock_physical_address[MAXTHREADS];

long lock_fail=0, lock_succeed=0;
long spec_rel_success=0, spec_rel_fail=0;

/* Because we don't want to recover from mispredicts and other such
   things until AFTER the cycle, we just record the event in these
   structures and do the recovery later */
scoreboard_t *mispredict_discovered[MAXTHREADS] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
scoreboard_t *lock_failed[MAXTHREADS] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
int misfetch_discovered[MAXTHREADS] = {0,0,0,0,0,0,0,0};

unsigned long flushed=0, execflushed=0;
extern unsigned long wpexec;

extern void
initsched()
{
  int i;
  for (i=0;i<MAXTHREADS;i++) {
    priority[i] = i;
  }
  for (i=0;i<MAXTHREADS;i++) {
    acquire_queue[i] = (acq_entry *)calloc(1, sizeof(acq_entry));
    acquire_queue[i]->addr = 0;
    acquire_queue[i]->timestamp = MAXLONG;
    acquire_queue[i]->valid = 0;
  }
}

static void clean_up_misfetch(context *current)
{
  int i;

  current->nextpc = current->rightpc;
  current->wrong_path = current->misfetching = 0;
  key[current->id] -= current->num_misfetches;
  if (current->cache_entry) {
    clean_up_cache_queue(current);
  }
  current->fetchcycle = cyc;
  current->wait_for_fetch = 0;
  current->num_misfetches = 0;
}

static void clean_up_mispredict(scoreboard_t *brinst) {

  context *current = thecontexts[brinst->thread];
  scoreboard_t *deadinst;
  int i;

  current->wrong_path = 0;
  current->nextpc = current->rightpc;
  DEBUGPRINTF("T%d fetch back on right path\n", current->id);
  if (current->cache_entry) {
    clean_up_cache_queue(current);
  }
  current->fetchcycle = cyc;
  current->wait_for_fetch = 0;
  bcopy(current->lasthazard_cp, current->lasthazard,MAXREG * sizeof(long));
  bcopy(current->writer_cp, current->writer, MAXREG * sizeof(int));
  bcopy(current->int_reg_cp, current->R,32 * sizeof(long));
  bcopy(current->fp_reg_cp, current->FPR,32 * sizeof(double));
  bcopy(current->return_stack_cp, current->return_stack, 12 * sizeof(long));
  current->rs_size = current->rs_size_cp;
  current->rs_start = current->rs_start_cp;
  current->ghr = current->ghr_cp;

  for (i = ((brinst->id+1) & (SCORESIZE-1));
       i != ((current->scoretop+1) & (SCORESIZE-1)) &&
       !(current->score[i].status & (INVALID | SQUASHED));
       i = (i+1) & (SCORESIZE-1)) {
    deadinst = &current->score[i];
    i_registers_freed += deadinst->iregs_used;
    f_registers_freed += deadinst->fregs_used;
    deadinst->deps = 0;
    deadinst->numwaiting = 0;
    deadinst->donecycle = cyc;
    deadinst->iregs_used = deadinst->fregs_used = 0;
#ifdef PRIO_INST_COUNT
    if (deadinst->status & FETCHED)
      key[brinst->thread]--;
#else
#ifdef PRIO_BR_COUNT
    if ((deadinst->status & FETCHED) && 
	(deadinst->branch || deadinst->cond_branch)) {
      key[brinst->thread]--;
    }
#endif
#endif
    if (deadinst->status != INVALID)
      deadinst->status = SQUASHED;
  }
}

static void
clean_up_lock(scoreboard_t *syncinst) {
  context *current = thecontexts[syncinst->thread];
  scoreboard_t *deadinst;
  int i, index;

  current->nextpc = syncinst->pc + 4;
  if (current->cache_entry) {
    clean_up_cache_queue(current);
  }
  current->fetchcycle = cyc;
  DEBUGPRINTF("T%d being flushed from queue\n", current->id);
  current->blocked = 1;
  bcopy(current->lasthazard_sync_cp,current->lasthazard,MAXREG * sizeof(long));
  bcopy(current->writer_sync_cp, current->writer, MAXREG * sizeof(int));
  bcopy(current->int_reg_sync_cp, current->R,32 * sizeof(long));
  bcopy(current->fp_reg_sync_cp, current->FPR,32 * sizeof(double));
  bcopy(current->return_stack_sync_cp,current->return_stack, 12*sizeof(long));
  current->rs_size = current->rs_size_sync_cp;
  current->rs_start = current->rs_start_sync_cp;
  current->ghr = current->ghr_sync_cp;

  for (i = ((current->restore_point + 1) & (SCORESIZE-1));
       i != ((current->scoretop+1) & (SCORESIZE-1));
       i = (i+1) & (SCORESIZE-1)) {
    deadinst = &current->score[i];
    i_registers_freed += deadinst->iregs_used;
    f_registers_freed += deadinst->fregs_used;
    deadinst->deps = 0;
    deadinst->numwaiting = 0;
    deadinst->donecycle = cyc;
    deadinst->iregs_used = deadinst->fregs_used = 0;
#ifdef PRIO_INST_COUNT
    if (deadinst->status & FETCHED)
      key[syncinst->thread]--;
#else
#ifdef PRIO_BR_COUNT
    if ((deadinst->status & FETCHED) && 
	(deadinst->branch || deadinst->cond_branch)) {
      key[syncinst->thread]--;
    }
#endif
#endif
    if (!(deadinst->status & FETCHED)) {
      execflushed++;
    }
    flushed++;
    current->instrs--;
    if (deadinst->status != INVALID)
      deadinst->status = SQUASHED;
  }
}


/* Look to see if any interesting mispeculations were detected
   that now need recovery action.
 */

void
fix_pcs() {
  int i, mispredict_found = 0;
  scoreboard_t *new, **prev;
  cache_request *cacherequest;

  for (i=0;i<numcontexts;i++) {
    if (misfetch_discovered[i]) {
      clean_up_misfetch(thecontexts[i]);
      misfetch_discovered[i] = 0;
      DEBUGPRINTF("T%d misfetch discovered\n", i);
    }
    else if (mispredict_discovered[i]) {
      mispredict_found = 1;
      clean_up_mispredict(mispredict_discovered[i]);
      mispredict_discovered[i] = NULL;
    }
    if (lock_failed[i]) {
      mispredict_found = 1;
      clean_up_lock(lock_failed[i]);
      lock_failed[i] = NULL;
    }
  }

  if (!mispredict_found)
    return;

  /* 
     take dead instructions out of queues 
       -brute force, inefficient.  but then, it works.
     */

  if (decode_head) {
    new = decode_head;
    prev = &(decode_head);
    decode_tail = NULL;
    while (new != NULL) {
      if (new->status & SQUASHED) {
	num_in_decode--;
      }
      else {
	decode_tail = *prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (rename_head) {
    new = rename_head;
    prev = &(rename_head);
    while (new != NULL) {
      if (new->status & SQUASHED) {
	num_in_rename--;
      }
      else {
	*prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (iq_head) {
    new = iq_head;
    prev = &(iq_head);
    iq_tail = NULL;
    while (new != NULL) {
      if (new->status & SQUASHED) {
	numiniq--;
      }
      else {
	iq_tail = *prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (fq_head) {
    new = fq_head;
    prev = &(fq_head);
    fq_tail = NULL;
    while (new != NULL) {
      if (new->status & SQUASHED) {
	numinfq--;
      }
      else {
	fq_tail = *prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (rread_head) {
    new = rread_head;
    prev = &(rread_head);
    rread_tail = NULL;
    while (new != NULL) {
      if (new->status & SQUASHED) {
	num_in_rread--;
      }
      else {
	rread_tail = *prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (exec_head) {
    new = exec_head;
    prev = &(exec_head);
    exec_tail = NULL;
    while (new != NULL) {
      if (new->status & SQUASHED) {
	num_in_exec--;
      }
      else {
	exec_tail = *prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }

  if (rwrite_head) {
    new = rwrite_head;
    prev = &(rwrite_head);
    while (new != NULL) {
      if (new->status & SQUASHED) {
	num_in_rwrite--;
      }
      else {
	*prev = new;
	prev = &(new->next);
      }
      new = new->next;
    }
    *prev = NULL;
  }
  
  cacherequest = cache_queue_head;
  while (cacherequest) {
    new = cacherequest->drequestor;
    prev = &(cacherequest->drequestor);
    while (new) {
      if (new->status & SQUASHED)
	*prev = new->mergeinst;
      else 
	prev = &(new->mergeinst);
      new = new->mergeinst;
    }
    cacherequest = cacherequest->next;
  }      
}


/* We checkpointed processor state at a lock instruction.  Now we
   are restarting the thread from that checkpointed state.  In this
   particular case, the lock in question succeeded.  Therefore, the
   speculative instructions beyond the lock must stay in the machine,
   but we have to re-execute them, now that we know they are not
   speculative, but also to make sure that any memory dependences
   protected by the lock are not violated.
 */

static void
restore_from_sync_cp(scoreboard_t *syncinst) {

  int next, index;
  scoreboard_t *nextinst;
  StashData *stash;
  unsigned long oldnextpc;

  context *current = thecontexts[syncinst->thread];
  current->blocked = 0;
  bcopy(current->lasthazard_sync_cp,current->lasthazard,MAXREG * sizeof(long));
  bcopy(current->writer_sync_cp, current->writer, MAXREG * sizeof(int));
  bcopy(current->int_reg_sync_cp, current->R,32 * sizeof(long));
  bcopy(current->fp_reg_sync_cp, current->FPR,32 * sizeof(double));
  bcopy(current->return_stack_sync_cp,current->return_stack, 12*sizeof(long));
  current->rs_size = current->rs_size_sync_cp;
  current->rs_start = current->rs_start_sync_cp;
  current->ghr = current->ghr_sync_cp;

  oldnextpc = current->nextpc;
  if (!syncinst->wp)
    current->wrong_path = 0;
  next = current->restore_point;
  if (current->nextsync == syncinst->id ) {
    /*FIXME this should always be true */
    current->nextsync = NONE;
    current->follow_sync = 0;
  }
  else printf("sync operations passing!\n");
  while (next != current->scoretop) {
    next = (next + 1) & (SCORESIZE - 1);
    nextinst = &current->score[next];
    if (!current->follow_sync)
      nextinst->wait_sync = 0;
    /* if (nextinst->status & EXECUTING)
      break; */
    if (current->wrong_path) {
      if (!nextinst->wp) {
	nextinst->wp = 1;
	current->wpinstrs++;
	current->instrs--;
	if (!(nextinst->status & FETCHED)) 
	  wpexec++;
      }
    }
    current->pc = nextinst->pc;
    DEBUGPRINTF("(resim)");
    SmtDISASSEMBLE(current->id, nextinst->pc, 1);
    current->nextpc = nextinst->pc + 4;
    if (!get_stash_data((mem_addr)nextinst->pc, current->id, &stash))
      decode_inst(current, stash);
    /*    if (!current->wrong_path)*/
      (*stash->emulate)(current,stash);
    if (current->srcmem || current->destmem) {
      nextinst->srcmem = current->srcmem;
      nextinst->destmem = current->destmem;
      current->srcmem = current->destmem = 0;
    }
    if (nextinst->status & (FETCHED | MEMORY)) {
      if (nextinst->dest < 31) {
	current->writer[nextinst->dest] = nextinst->id;
      }
      if (nextinst->dest > 31 && nextinst->dest != 63) {
	current->writer[nextinst->dest] = nextinst->id;
      }
    }
    if (!current->wrong_path && !current->follow_sync
	&& (nextinst->syncop == SMT_HW_LOCK || nextinst->syncop == LDL_L
	    || nextinst->syncop == LDQ_L || nextinst->syncop == STL_C
	    || nextinst->syncop == STQ_C || nextinst->syncop == READ_FULL
	    || nextinst->syncop == WRITE_FILL)) {
      if (nextinst->syncop == WRITE_FILL) {
	nextinst->fer_value = current->FPR[nextinst->src1-32];
      }
      current->nextsync = nextinst->id;
      current->follow_sync = 1;
      nextinst->pc = current->pc;
	  
      DEBUGPRINTF("T%d in follow_sync mode\n", current->id);
      bcopy(current->lasthazard, current->lasthazard_sync_cp, MAXREG * sizeof(long));
      bcopy(current->writer, current->writer_sync_cp, MAXREG * sizeof(int));
      bcopy(current->R, current->int_reg_sync_cp,32 * sizeof(long));
      bcopy(current->FPR, current->fp_reg_sync_cp,32 * sizeof(double));
      bcopy(current->return_stack, current->return_stack_sync_cp,
	    12 * sizeof(long));
      current->rs_size_sync_cp = current->rs_size;
      current->rs_start_sync_cp = current->rs_start;
      current->ghr_sync_cp = current->ghr;
      current->restore_point = nextinst->id;
    }
    if (nextinst->branch && !current->follow_sync && !current->wrong_path) {
      if (nextinst->condbranch) {
	current->misfetch = 1 - btblookup(current->pc, 0,0,
					  current->masterid,
					  stash->taken, BRANCH);
	current->prediction = bpredict(current->pc, stash->taken,
				       current->ghr);
	nextinst->ghr = current->ghr;
	current->ghr = ((current->ghr << 1) | stash->taken) & PHTMASK; 
      }
      if (current->prediction == PREDICT_WRONG  || current->misfetch
	  || (nextinst->taken != stash->taken)) {
	DEBUGPRINTF("T%d mispredicting (after sync)\n", current->id);
	if (nextinst->condbranch) {
	  if (stash->taken) {
	    current->rightpc = stash->br_target;
	  }
	  else {
	    current->rightpc = current->pc + 4;
	 }
	  if (current->misfetch) {
	    current->nextpc = current->pc + 4;
	  }
	}
	else if (!current->misfetch && nextinst->taken != stash->taken) {
	  /* unconditional branch */
	  current->rightpc = current->nextpc;
	  current->misfetch = 1;
	}
	if ((current->prediction == PREDICT_CORRECT && current->misfetch)
	    || !(nextinst->status & (FETCHED | EXECUTING))) {
	  mispredict_discovered[current->id] = nextinst;
	}
	else {
	  nextinst->mispredict = 1;
	}
	nextinst->taken = stash->taken;
	current->misfetch = 0;
	current->wrong_path = 1;
	current->prediction = PREDICT_CORRECT;
	bcopy(current->lasthazard, current->lasthazard_cp,MAXREG * sizeof(long));
	bcopy(current->writer, current->writer_cp, MAXREG * sizeof(int));
	bcopy(current->R, current->int_reg_cp,32 * sizeof(long));
	bcopy(current->FPR, current->fp_reg_cp,32 * sizeof(double));
	bcopy(current->return_stack, current->return_stack_cp,
	      12 * sizeof(long));
	current->rs_size_cp = current->rs_size;
	current->rs_start_cp = current->rs_start;
	current->ghr_cp = current->ghr;
      }
    }
  }
  current->nextpc = oldnextpc;
}


/* We checkpointed processor state at a lock instruction.  Now we
   are restarting the thread from that checkpointed state.
 */

static void
restart_thread(scoreboard_t *syncinst) {

  int next, index;
  scoreboard_t *nextinst;
  StashData *stash;
  unsigned long oldnextpc;

  context *current = thecontexts[syncinst->thread];
  current->blocked = 0;
  current->fetchcycle = cyc + 1;

  bcopy(current->lasthazard_sync_cp,current->lasthazard,MAXREG * sizeof(long));
  bcopy(current->writer_sync_cp, current->writer, MAXREG * sizeof(int));
  bcopy(current->int_reg_sync_cp, current->R,32 * sizeof(long));
  bcopy(current->fp_reg_sync_cp, current->FPR,32 * sizeof(double));
  bcopy(current->return_stack_sync_cp,current->return_stack, 12*sizeof(long));
  current->rs_size = current->rs_size_sync_cp;
  current->rs_start = current->rs_start_sync_cp;
  current->ghr = current->ghr_sync_cp;
  syncinst->donecycle = cyc;
}

long prev_lock_address[MAXTHREADS][1024];

/* Lock release speculation -- see HPCA99 */
void
check_early_release(scoreboard_t *top) {
  int i, oldest_id;
  int nextid = (top->thread + 1) % numcontexts;
  context *nextthread = thecontexts[nextid];
  long oldest_stamp;

  if (top->syncop == SMT_RELEASE) {
    long prev_lock = prev_lock_address[top->thread][(top->pc>>2) & 1023];
    oldest_id = NONE;
    oldest_stamp = cyc+1;
    for (i=0;i<numcontexts;i++) {
      if (acquire_queue[i]->valid && acquire_queue[i]->addr == prev_lock &&
	  acquire_queue[i]->timestamp < oldest_stamp) {
	oldest_id = i;
	oldest_stamp = acquire_queue[i]->timestamp;
      }
    }
    if (oldest_id != NONE) {
      nextthread = thecontexts[oldest_id];
      if (nextthread->score[nextthread->scoretop].status == INVALID) {
#ifdef DEBUG
	if (debugsync)
	  printf("Context %d doing a speculative release of context %d\n",
		 top->thread, oldest_id);
#endif
	restart_thread(acquire_queue[oldest_id]->inst);
	nextthread->restore_point = nextthread->scoretop;
	thecontexts[top->thread]->early_release_thread = oldest_id;
	thecontexts[top->thread]->early_release_inst = top->id;
      }
    }
  }
  else if ((top->syncop == WRITE_FILL && top->fer_thread == nextid &&
	    nextthread->blocked && 
	    nextthread->wait_on_full[top->fer_reg])){
    /*this next if will fail if the lock fail and the release fetch
      occur the same cycle */
    if (nextthread->score[nextthread->scoretop].status == INVALID) {
#ifdef DEBUG
      if (debugsync)
	printf("Context %d doing a speculative release of context %d\n",
	       top->thread, nextid);
#endif
      restart_thread(nextthread->wait_on_full[top->fer_reg]);
      nextthread->restore_point = nextthread->scoretop;
      thecontexts[top->thread]->early_release_thread = nextid;
      thecontexts[top->thread]->early_release_inst = top->id;
      thecontexts[top->thread]->early_release_reg = top->src1;
    }
  }
}


/* Lock instructions are not emulated in fetch, like everything else.
   They must execute in processor order, so they are emulated here.
   But even this is deceptive, because this routine is typically called
   from the commit stage, rather than the execute stage.  That is because
   I typically assume synch instructions don't take effect until they
   retire.  yes, that's pretty conservative.
 */

synchexecute(scoreboard_t *exinst) {
  unsigned long va;
  unsigned long least;
  int leastid, releaser, releasee;
  int i;
  context *current;
  int num_waiters;
  scoreboard_t *waitinst;

  current = thecontexts[exinst->thread];
  if (exinst->syncop) {
    if (!exinst->wp) {
      switch(exinst->syncop) {
      case SMT_HW_LOCK:
	va = exinst->srcmem;
	if ((signed int) read_mem_word(current, va)) {
	  /* someone else is holding the lock */
#ifdef DEBUG
	  if (debugsync) 
	    printf("\t\tContext %d trying HW_LOCK 0x%lx...someone else holding lock, cyc=%ld, pc=0x%lx\n", current->id, va, cyc, exinst->pc);
#endif
	  lock_fail++;
	  acquire_queue[current->id]->addr = va;
	  acquire_queue[current->id]->valid = 1;
	  acquire_queue[current->id]->timestamp = cyc;
	  acquire_queue[current->id]->inst = exinst;
	  lock_failed[current->id] = exinst;
	  exinst->donecycle = cyc + 100000;
	} else {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d trying HW_LOCK 0x%lx...obtained, cyc=%ld, pc = 0x%lx\n", current->id, va, cyc, exinst->pc);
#endif
	  lock_succeed++;
	  acquire_queue[current->id]->valid = 0;
	  set_mem_word(current, va, 0x1);
	  restore_from_sync_cp(exinst);
	}
	break;
      case SMT_RELEASE:
	va = exinst->destmem;
	/* Check if anyone else is waiting for the lock */
	least = MAXLONG;
	leastid = -1;
	prev_lock_address[exinst->thread][(exinst->pc>>2) & 1023] = va;
	for (i=0; i < numcontexts; i++) {
	  if (acquire_queue[i]->valid && acquire_queue[i]->addr == va) {
	    if (acquire_queue[i]->timestamp < least) {
	      least = acquire_queue[i]->timestamp;
	      leastid = i;
	    }
	  }
	}
	if (leastid >= 0) {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d SMT_RELEASEing 0x%lx...context %d waiting, cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, leastid, cyc, exinst->pc);
#endif

	  acquire_queue[leastid]->valid = 0;
	  if (lock_failed[leastid] == acquire_queue[leastid]->inst) {
	    lock_failed[leastid] = NULL;
	    restore_from_sync_cp(acquire_queue[leastid]->inst);
	  }
	  else {
	    if (!thecontexts[leastid]->blocked) {
	      spec_rel_success++;
	      restore_from_sync_cp(acquire_queue[leastid]->inst);
	    }
	    else {
	      spec_rel_fail++;
	      restart_thread(acquire_queue[leastid]->inst);
	      thecontexts[leastid]->follow_sync = 0;
	      thecontexts[leastid]->nextsync = NONE;
	    }
	  }
	  if (current->early_release_thread != NONE &&
	      current->early_release_thread != leastid) {
#ifdef DEBUG
	    if (debugsync)
	      printf("Speculative release of context %d by context %d aborted\n",
		     current->early_release_thread,exinst->thread);
#endif
	    lock_failed[current->early_release_thread] = 
	      acquire_queue[current->early_release_thread]->inst;
	    releaser = current->early_release_thread;
	    releasee = thecontexts[releaser]->early_release_thread;
	    thecontexts[releaser]->early_release_thread = NONE;
	    while (releasee != NONE && releasee != leastid) {
	      if (lock_failed[releasee] != 0 ||          /*already aborted */
		  !acquire_queue[releasee]->valid) { /* stale information */
		break;
	      }
#ifdef DEBUG
	      if (debugsync)
		printf("Speculative release of context %d by context %d aborted\n",
		       releasee,releaser);
#endif
	      lock_failed[releasee] = 
		acquire_queue[releasee]->inst;
	      releaser = releasee;
	      releasee = thecontexts[releaser]->early_release_thread;
	      thecontexts[releaser]->early_release_thread = NONE;
	    }
	  }
	  current->early_release_thread 
	    = current->early_release_inst = NONE;
	} else {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d SMT_RELEASEing 0x%lx...no one waiting, cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, cyc, exinst->pc, current->R[26]);
#endif
	  current = thecontexts[exinst->thread];
	  set_mem_word(current, va, 0x0);
	  for (i=0; i<MAXTHREADS; i++) {
	    if (lock_physical_address[i] == exinst->destmem 
		&& i != exinst->thread) {
	      lock_flag[i] = 0;
	    }
	  }
	}
	break;	  
      case SMT_TERMINATE:
#ifdef DEBUG
	if (debugsync)
	  printf("SMT_terminate (context %d), cyc = %d, pc = 0x%lx\n", 
		 current->id, cyc, exinst->pc);
#endif
	break;
      case LDL_L:
      case LDQ_L:
	va = exinst->srcmem;
	lock_flag[current->id] = 1;
	lock_physical_address[current->id] = va;
	(signed long) current->int_reg_sync_cp[exinst->dest] = 
	  (signed int) read_mem_word(current, va);
#ifdef DEBUG
	if (debugsync)
	  printf("\t\tContext %d LDL_Locking 0x%lx read %d ...cyc = %ld, pc=0x%lx\n", 
		 exinst->thread, va, current->int_reg_sync_cp[exinst->dest],
		 cyc, exinst->pc);
#endif
	restore_from_sync_cp(exinst);
	break;
      case STL_C:
      case STQ_C:
	va = exinst->destmem;
	if (lock_flag[current->id]) {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d STL_C success 0x%lx writing %d ...cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, current->int_reg_sync_cp[exinst->dest],
		   cyc, exinst->pc);
#endif
	  set_mem_word(current, va,
		       current->int_reg_sync_cp[exinst->src2] & 
		       (unsigned long) 0xffffffff);
	  current->int_reg_sync_cp[exinst->dest] = 1;
	  for (i=0;i<MAXTHREADS;i++) {
	    if (lock_physical_address[i] == va) {
	      lock_flag[i] = 0;
	    }
	  }
	}
	else {
	  current->int_reg_sync_cp[exinst->dest] = 0;
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d STL_C fail 0x%lx...cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, cyc, exinst->pc);
#endif
	}
	lock_physical_address[current->id] = 0;
	lock_flag[current->id] = 0;
	for (i=0;i<exinst->numwaiting;i++) {
	  waitinst = exinst->waiter[i];
	  exinst->waiter[i] = 0;
	  if (waitinst->hazcycle < exinst->donecycle) {
	    waitinst->hazcycle = exinst->donecycle;
	  }
	  if (--waitinst->deps == 0) {
	    waitinst->status &= ~BLOCKED;
	    waitinst->readycycle = waitinst->hazcycle;
	  }
	}
	exinst->numwaiting = 0;
	if (current->writer_sync_cp[exinst->dest] == exinst->id) {
	  current->writer_sync_cp[exinst->dest] = NONE;
	  current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
	}
	restore_from_sync_cp(exinst);
	break;

/*

	va = exinst->destmem;
	if (lock_flag[current->id]) {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d STL_C success 0x%lx writing %d ...cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, current->R[exinst->dest],
		   cyc, exinst->pc);
#endif
	  set_mem_word(current, va,
		       current->R[exinst->src2] & (unsigned long) 0xffffffff);
	  current->R[exinst->dest] = 1;
	  for (i=0;i<MAXTHREADS;i++) {
	    if (lock_physical_address[i] == va) {
	      lock_flag[i] = 0;
	    }
	  }
	}
	else {
	  current->R[exinst->dest] = 0;
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d STL_C fail 0x%lx...cyc = %ld, pc=0x%lx\n", 
		   exinst->thread, va, cyc, exinst->pc);
#endif
	}
	lock_physical_address[current->id] = 0;
	lock_flag[current->id] = 0;
*/
      case SMT_TAS:
	/* FIXME -- not right yet */
#ifdef MFLOW
	va = current->R[2];
	current->R[2] = (signed int) read_mem_word(current, va);
#else
	va = current->R[REG_A0];
	current->R[REG_V0] = (signed int) read_mem_word(current, va);
#endif
	set_mem_word(current, va, 0x1);
	/* release(exinst);*/
	DEBUGPRINTF("SMT_TAS (context %d), addr = 0x%lx, prev value = %d, cyc = %d, pc = 0x%lx\n", 
		    current->id, va, current->R[2], cyc, current->R[26]);
	break;
      case WRITE_FILL: /* move to other reg file */
	if (thecontexts[exinst->fer_thread]->full[exinst->fer_reg] == EMPTY) {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d WRITING %lf into thread %d reg %d, cyc=%ld, pc=0x%lx\n", 
		   current->id, exinst->fer_value, exinst->fer_thread, 
		   exinst->fer_reg, cyc, exinst->pc);
#endif
	  thecontexts[exinst->fer_thread]->FPR[exinst->fer_reg] = 
	    thecontexts[exinst->fer_thread]->fp_reg_cp[exinst->fer_reg] = 
	    thecontexts[exinst->fer_thread]->fp_reg_sync_cp[exinst->fer_reg] = 
	    exinst->fer_value;
	  if (thecontexts[exinst->fer_thread]->wait_on_full[exinst->fer_reg]) {
	    if (thecontexts[exinst->fer_thread]->blocked) {
	      if (current->early_release_thread != NONE) {
#ifdef DEBUG
		if (debugsync)
		  printf("Speculative release of context %d by context %d aborted\n",
			 current->early_release_thread,exinst->thread);
#endif
		lock_failed[current->early_release_thread] = 
		  thecontexts[current->early_release_thread]->wait_on_full[current->early_release_reg];
		current->early_release_inst = NONE;
		current->early_release_thread = NONE;
		current->early_release_reg = NONE;
	      }
	      if (lock_failed[exinst->fer_thread] && 
		  lock_failed[exinst->fer_thread]->dest == exinst->src1) {
		/* blocked this cycle, hasn't been flushed yet */
		thecontexts[exinst->fer_thread]->blocked = 0;
		restore_from_sync_cp(lock_failed[exinst->fer_thread]);
		lock_failed[exinst->fer_thread] = NULL;
	      }
	      else {
		restart_thread(thecontexts[exinst->fer_thread]->wait_on_full[exinst->fer_reg]);
		thecontexts[exinst->fer_thread]->follow_sync = 0;
		thecontexts[exinst->fer_thread]->nextsync = NONE;
	      }
	      thecontexts[exinst->fer_thread]->writer[exinst->fer_reg] = NONE;
	      thecontexts[exinst->fer_thread]->lasthazard[exinst->fer_reg] 
		= exinst->donecycle;
	      thecontexts[exinst->fer_thread]->wait_on_full[exinst->fer_reg] = 0;
	    }
	    else if (current->early_release_thread == exinst->fer_thread) {
	      thecontexts[exinst->fer_thread]->full[exinst->fer_reg] = FULL;
#ifdef DEBUG
	      if (debugsync)
		printf("(re-exec)");
#endif
	      synchexecute(thecontexts[exinst->fer_thread]->wait_on_full[current->early_release_reg]);
	      thecontexts[exinst->fer_thread]->wait_on_full[current->early_release_reg] = NULL;
	      current->early_release_thread 
		= current->early_release_inst 
		  = current->early_release_reg = NONE;
	    }
	    else printf("write_fill did nothing\n");
	  }
	  else {
	    thecontexts[exinst->fer_thread]->full[exinst->fer_reg] = FULL;
	  }
	  exinst->donecycle = cyc + exinst->delay;
	  restore_from_sync_cp(exinst);
	}
	else { /* FIXME should block on write_full if already full,
		  not implemented yet */
	}
	break;
      case READ_FULL: /* read register if full */
	if (current->full[exinst->dest]) {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d READ_IF_FULL from reg %d reads %lf, cyc=%ld, pc=0x%lx\n",
		   current->id, exinst->dest, 
		   current->int_reg_sync_cp[exinst->dest], cyc, exinst->pc);
#endif
	  current->full[exinst->dest] = EMPTY;
	  if (current->writer_sync_cp[exinst->dest] == exinst->id) {
	    current->writer_sync_cp[exinst->dest] = NONE;
	    current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
	  }
	  restore_from_sync_cp(exinst);
	}
	else {
#ifdef DEBUG
	  if (debugsync)
	    printf("\t\tContext %d READ_IF_FULL reg %d MUST WAIT, cyc=%ld, pc=0x%lx\n", current->id, exinst->dest, cyc, current->pc);
#endif
	  current->wait_on_full[exinst->dest] = exinst;
	  lock_failed[exinst->thread] = exinst;
	  current->blocked = 1;
	}
	exinst->donecycle = cyc + exinst->delay;
	for (i=0;i<exinst->numwaiting;i++) {
	  waitinst = exinst->waiter[i];
	  exinst->waiter[i] = 0;
	  if (waitinst->hazcycle < exinst->donecycle) {
	    waitinst->hazcycle = exinst->donecycle;
	  }
	  if (--waitinst->deps == 0) {
	    waitinst->status &= ~BLOCKED;
	    waitinst->readycycle = waitinst->hazcycle;
	  }
	}
	exinst->numwaiting = 0;
	break;
      }
    }
  }
}


/* It's okay to put them in this queue backwards, because I'll
   flip them around again when I take them out.
 */

static void
insert_in_rwrite(scoreboard_t *newinst)
{
  num_in_rwrite++;
  newinst->next = rwrite_head;
  rwrite_head = newinst;
}


/* for most instructions, not much interesting happens in the
   execute stage, because dependence resolution happens in the
   queue stage.

   This is, however, where I detect mispredictions.
 */

extern void
execute()
{
  scoreboard_t *instrn, *nextinstrn, **prev = &exec_head;

  instrn = exec_head;
  exec_head = exec_tail = NULL;
  while (instrn) 
    {
      if (instrn->donecycle <= cyc - 1) {
	DEBUGPRINTF("T%ds%d completes execution %s, delay %d (%d)\n",
		    instrn->thread, instrn->id, instrn->wp ? "(wp)":"",
		    instrn->delay,instrn->donecycle+2);
	if (instrn->mispredict) {
	  mispredict_discovered[instrn->thread] = instrn;
	  DEBUGPRINTF("T%d mispredict discovered\n", instrn->thread);
	}
	if (instrn->status == EXECUTING)
	  instrn->status = EXECUTED; /* ensures that mispredict not recorded
					twice, in particular by
					restore_from_sync_cp */
	if (instrn->fu == SYNCH && 
	    (instrn->syncop == LDL_L || instrn->syncop == LDQ_L)) {
	  synchexecute(instrn);
	}
	nextinstrn = instrn->next;
	insert_in_rwrite(instrn);
	instrn = nextinstrn;
	num_in_exec--;
      }
      else {
	exec_tail = instrn;
	*prev = instrn;
	prev = &(instrn->next);
	instrn = instrn->next;
      }
    }
}
