/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */


/* Alpha Instructions */
typedef enum {
  CALL_PAL,LDA=8,LDAH,LDQ_U=0xb,STQ_U=0xf,
  INTA,INTL,INTS,INTM,FLTV=0x15,FLTI,FLTL,MISC,JMPJSR=0x1a,
  LDF=0x20,LDG,LDS,LDT,STF,STG,STS,STT,LDL,LDQ,LDL_L,LDQ_L,STL,STQ,STL_C,STQ_C,
  BR,FBEQ,FBLT,FBLE,BSR,FBNE,FBGE,FBGT,BLBC,BEQ,BLT,BLE,BLBS,BNE,BGE,BGT
} opcodefld;

typedef enum { /*mem functions, opcode = 18*/
    FETCH = 0x8000,
    RC = 0xE0000,
    TRAPB = 0x0,
    FETCH_M = 0xA000,
    RPCC = 0xC000,
    MB = 0x4000,
    RS = 0xF000 } memfield;

typedef enum {  /* jump instructions, opcode = 1a*/
  JMP, JSR, RETN, JSR_COROUTINE } jmpfield;

typedef enum { /*operate, opcode = 10 */
  ADDL=0,ADDQV=0x60,CMPLE=0x6D,CMPULT=0x1D,SUBQ=0x29,S4ADDL=0x2,S4SUBQ=0x2B,
  S8SUBL=0x1B,ADDLV=0x40,CMPBGE=0xf,CMPLT=0x4d,SUBL=0x9,SUBQV=0x69,S4ADDQ=0x22,
  S8ADDL=0x12,S8SUBQ=0x3B,ADDQ=0x20,CMPEQ=0x2D,CMPULE=0x3D,SUBLV=0x49,
  S4SUBL=0xb,S8ADDQ=0x32} op10field;

typedef enum { /* operate, opcode = 11 */
  AND=0,BIC=0x8,BIS=0x20,CMOVEQ=0x24,CMOVLBC=0x16,CMOVLBS=0x14,CMOVGE=0x46,
  CMOVGT=0x66,CMOVLE=0x64,CMOVLT=0x44,CMOVNE=0x26,EQV=0x48,ORNOT=0x28,XOR=0x40
  } op11field;

typedef enum { /* operate, opcode = 12 */
  EXTBL=0x06,EXTLH=0x6a,EXTLL=0x26,EXTQH=0x7a,EXTQL=0x36,EXTWH=0x5a,
  EXTWL=0x16,INSBL=0x0b,INSLH=0x67,INSLL=0x2b,INSQH=0x77,INSQL=0x3b,
  INSWH=0x57,INSWL=0x1b,MSKBL=0x2,MSKLH=0x62,MSKLL=0x22,MSKQH=0x72,
  MSKQL=0x32,MSKWH=0x52,MSKWL=0x12,SLL=0x39,SRA=0x3C,SRL=0x34,
  ZAP=0x30,ZAPNOT=0x31} op12field;

typedef enum { /* operate, opcode = 13 */
  MULL=0,MULLV=0x40,MULQ=0x20,MULQV=0x60,UMULH=0x30} op13field;

typedef enum { /* float, opcode = 17 */
  CPYS=0x20,CPYSE=0x22,CPYSN=0x21,CVTLQ=0x10,CVTQL=0x30,CVTQLSV=0x530,
  CVTQLV=0x130,FCMOVEQ=0x2a,FCMOVGE=0x2d,FCMOVGT=0x2f,FCMOVLE=0x2e,
  FCMOVLT=0x2c,FCMOVNE=0x2b,MF_FPCR=0x25,MT_FPCR=0x24} float17field;

typedef enum { /* float, opcode = FLTI */
  ADDS=0,ADDT=0x20,CMPTEQ=0x25,CMPTLT=0x26,CMPTLE=0x27,CMPTUN=0x24,
  CVTQS=0x3c,CVTQT=0x3e,CVTTS=0x2c,DIVS=0x3,DIVT=0x23,MULS=0x2,MULT=0x22,
  SUBS=0x1,SUBT=0x21,CVTTQ=0x2f} floatop;

typedef enum { /*palcode, opcode = 00 */
  callsys=0x83, rduniq=0x9e, wruniq=0x9f } palop;

