/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdlib.h>
#include <string.h>
#include "main.h"
#include "mem.h"
#include "predict.h"

extern int debugsync;
int systrace = 0, debug = 0;
jmp_buf toplevel;
static char **env;
int spec_release = 0;
int spec_fill = 0;

int numcontexts;
context *thecontexts[MAXTHREADS];

static void
init_threads(char *env[], char *filename, int *curcontext)
{

  FILE *argfile;
  int retcode,i,numthreads,cont;
  int curthread = 0;
  int locargc = 0;
  char *locargv[20] = {0,0,0,0,0,0,0,0,0,0};
  char file[256];
  FILE *infile = NULL, *outfile = NULL;
  context *current, *master;

  current = thecontexts[*curcontext] = (context *) calloc(1,sizeof(context));
  current->argfile = argfile = fopen(filename, "r");
  locargv[0] = (char *) malloc(240);

  /*
   ** Skip first symbol - assumed to be a '#' character
   */
  retcode = fscanf(argfile, "%s", locargv[0]);
  fscanf(argfile, "%d", &numthreads);

  while (1) {
    if (locargv[locargc] == NULL) {
      locargv[locargc] = (char *) malloc(240);
    }

    /*
    ** Read until EOF or the next '#' symbol
    */
    retcode = fscanf(argfile, "%s", locargv[locargc]);
    if (retcode == EOF || locargv[locargc][0] == '#') {
      current->FWR = (void *) &current->FPR[0];
      current->lasthazard[31] = -1;
      current->scoretop = 0;
      for (i=0;i<MAXREG;i++) {
	current->writer[i] = NONE;
      }
      current->id = *curcontext;
      current->prediction = PREDICT_CORRECT;
      if(readaout(current, locargv[0],(context *) NULL, 0, numthreads) != 0)
	error(Fatal,"Cannot open input file\n");
      printf("thread %d %s\n", *curcontext, locargv[0]);
      setupstack(current, 0, locargc, locargv, env);
      locargc = 0;
      current->nextpc = current->pc = current->pstaddress;
      current->running = 1;
      current->next_to_commit = 1;
      current->infile = infile;
      current->outfile = outfile;
      current->nextsync = NONE;
      current->early_release_inst = NONE;
      current->early_release_thread = NONE;
      current->fetchcycle = (*curcontext)++ * 60;
      if (retcode == EOF) current->argfile = NULL;
      for (i=0;i<SCORESIZE;i++) {
	current->score[i].id = i;
	current->score[i].status = INVALID;
	current->score[i].deps = 0;
      }
      current->masterid = current->id;
      master = current;
      for (cont=1;cont<numthreads;cont++) {
	current = thecontexts[(*curcontext)] = (context *) calloc(1,sizeof(context));
	current->argfile = argfile = fopen(filename, "r");
	current->FWR = (void *) &current->FPR[0];
	current->lasthazard[31] = -1;
	current->scoretop = 0;
	for (i=0;i<MAXREG;i++) {
	  current->writer[i] = NONE;
	}
	current->id = *curcontext;
	current->prediction = PREDICT_CORRECT;
	printf("thread %d %s\n", *curcontext, locargv[0]);
	readaout(current, locargv[0], master, 1, numthreads);
	setupstack(current, 0, locargc, locargv, env);
	locargc = 0;
	current->running = 0;
	current->next_to_commit = 1;
	current->infile = infile;
	current->outfile = outfile;
	current->nextsync = NONE;
	current->early_release_inst = NONE;
	current->early_release_thread = NONE;
	current->fetchcycle = (*curcontext)++ * 60;
	if (retcode == EOF) current->argfile = NULL;
	current->masterid = master->id;
	for (i=0;i<SCORESIZE;i++) {
	  current->score[i].id = i;
	  current->score[i].status = INVALID;
	  current->score[i].deps = 0;
	}
	
      }
      for (i=0;i<locargc;i++)
	free(locargv[i]);
      return;
    }
    else if (locargv[locargc][0] == '<') {
      fscanf(argfile, "%s", file);
      infile = fopen(file,"r");
      if (infile == NULL) {
	printf("couldn't open %s for reading\n",file);
	exit(1);
      }
    }
    else if (locargv[locargc][0] == '>') {
      fscanf(argfile, "%s", file);
      outfile = fopen(file,"w");
      if (outfile == NULL) {
	printf("couldn't open %s for writing\n",file);
	exit(1);
      }
    }
    else {
      locargc++;
      }
  }
}
	
void
main(int argc, char *argv[], char *mainenv[])
{
    static int i = 1, j, k, oldi, exitcode;
    char *env_ptr = NULL;
    long ffdist[MAXTHREADS];

#ifdef __alpha
#include <sys/sysinfo.h>
#include <sys/proc.h>
    /*
    ** In the case of unaligned accesses on alpha machines: suppress
    ** the printing of error messages and cause a SIGBUS signal to be
    ** delivered to the thread.
    */
    int buf[2] = { SSIN_UACPROC, UAC_NOPRINT | UAC_SIGBUS};
    if(setsysinfo((long)SSI_NVPAIRS, (caddr_t)buf, 1L, 0L, 0L) < 0) {
	printf("setsysinfo failed\n");
	exit(2);
    }
#endif

    /*
    ** Handle command line variables.
    */

/*  The command line options are these:
 *    -d   -- debug mode -- prints a LOT of output
 *    -systrace  -- traces system calls
 *    -dsync -- debugs synchronization actions
 *    -ff dist -- sets the fast-forward distance for all threads
 *    -f thread dist -- sets the fast-forward distance for a single thread
 */

    for (k=0;k<MAXTHREADS;k++)
      ffdist[k] = 0;

    env = mainenv;
    if(argc - i > 0)
	do {
	    oldi = i;
	    if(strcmp("-systrace", argv[i]) == 0)
		systrace = 1,  ++i;
	    else if(strcmp("-d", argv[i]) == 0)
		debug = 1, ++i;
	    else if(strcmp("-dsync", argv[i]) == 0)
		debugsync = 1, ++i;
	    else if(strcmp("-specrel", argv[i]) == 0)
		spec_release = 1, ++i;
	    else if(strcmp("-specfill", argv[i]) == 0)
		spec_fill = 1, ++i;
	    else if(strcmp("-ff", argv[i]) == 0) {
	      i++;
	      for (k=0;k<MAXTHREADS;k++) {
		ffdist[k] = atol(argv[i]);
	      }
	      i++;
	    }
	    else if(strcmp("-f", argv[i]) == 0) {
	      ffdist[ atoi(argv[++i])] = atol(argv[++i]);
	      i++;
	    }
	    else break; /*rest of args should be filenames*/
	} while((argc - i > 0) && (oldi != i));

    if(setjmp(toplevel) == 0)
    {
        initfds();
	initcache();
	init_pht(); 
	init_btb();
	init_tlbs();
	initsched();
	init_stack_space();
	numcontexts = 0;

	/*
	** Initialize active threads that have immediate work to do.
	*/
	for (j=i;(j<argc) && (numcontexts < MAXTHREADS);j++) {
	  DEBUGPRINTF("Initializing context %d\n", numcontexts);
	  init_threads(env, argv[j], &numcontexts);
	}


#ifdef SMT_DEBUG2
	print_memstats();
#endif

	sigignore(SIGFPE);

/*
 *  fast-forward  skips the first (ffdist[]) instructions
 *      before beginning simulation.  
 */

	fast_forward(ffdist);
	zero_pstats();

	/*
	** Run the threads
	*/
	exitcode = run(env) - 256;

	if(exitcode > 0)
	  printf("Program exited with status %d.\n", exitcode);
	exit(exitcode);
    }
}
