/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#ifndef _MAIN_H_DEFINED
#define _MAIN_H_DEFINED

#include <setjmp.h>			/* Need the 'jmp_buf' type	*/
#include <stdio.h>			/* Need the 'FILE' type		*/
#include "stash.h"
#include "predict.h"

#define PRIO_INST_COUNT

#ifdef DEBUG
#define DEBUGPRINTF(fmt,args...)	 if (debug) printf(fmt, ## args)
#define SmtDISASSEMBLE(x, y, z)	if(debug) print(x, y, z);
#else
#define DEBUGPRINTF(fmt,args...)		
#define SmtDISASSEMBLE(x, y, z)	;
#endif

#define MAXTHREADS 8
#define MAXFDS      64
#define PAGESIZE   8192 /* pagesize */
#define PAGEMASK  (~(PAGESIZE-1)) /* Mask out the offset with in a page */
#define MAXREG     64

#define NONE -1


typedef enum {FP, SYNCH, INTEGER, INTLDST} futype;
typedef enum {FETCHED=1, BLOCKED=2, FETCHED_BLOCKED=3, EXECUTING=4,
	      INVALID=8, RETIREABLE=16, SQUASHED=32, MEMORY=64,
	      EXECUTED=128} execstatus;


/* The scoreboard (poor name--this really is analagous to the
   active list in MIPS/Alpha or reorder buffer in others.
   This is the mechanism that keeps track of each instruction
   that is in flight for a particular thread.
 */
typedef struct scoreboard{
  futype fu;
  execstatus status;
  int src1, src2, dest, deps;
  int thread, id, branch, condbranch, taken, ghr;
  int delay;
  int syncop, wait_sync;
  long srcmem, destmem;
  int numwaiting;
  struct scoreboard *waiter[50];
  unsigned long readycycle;
  unsigned long hazcycle;
  unsigned long donecycle;
  int iregs_used, fregs_used;
  int mispredict;
  int misfetch;
  int wp;
  int noop;
  int fer_thread;
  int fer_reg;
  double fer_value;
  unsigned long pc;
  struct scoreboard *next;
  struct scoreboard *mergeinst;
} scoreboard_t;

/* If you want to actually simulate the performance of
   a limited-size reorder buffer, change this number.
   Otherwise, I keep it high.
 */

#define SCORESIZE 256

typedef unsigned long int mem_addr;
typedef unsigned long mem_quad;
typedef unsigned int mem_word;
typedef unsigned short int mem_half;
typedef unsigned char mem_byte;

typedef enum { Text, Data, Stack } MemId;

typedef enum { EMPTY, FULL} full_empty;

struct segment_t {
    MemId segmentid;
    mem_addr base;      /* starting address in virtual memory */
    unsigned long size; /* size of the segment */
    void *segment;
};
typedef struct segment_t Segment;

/* This is everything you might possibly need to know about
   a hardware thread (context), which is quite a lot.  But doing
   things in this way allows me to do a context switch by just
   changing the pointer to the current context, and accessing
   everything indirectly through that pointer (typically called
   current).
 */

typedef struct context_t {
 
  int running;
  int blocked;
  /* declare machine state and registers */
  unsigned long R[32]; /* CPU registers */
  double FPR[32];      /* FPU registers */
  unsigned long int_reg_cp[32];
  double fp_reg_cp[32];
  unsigned long int_reg_sync_cp[32];
  double fp_reg_sync_cp[32];
  full_empty full[64];
  scoreboard_t *wait_on_full[64];
  long *FWR;
  long fpcr;
  unsigned long pc, nextpc;
  unsigned long hwpcb;
  broutcome prediction;
  int nextsync;
  int follow_sync;
  int early_release_thread;
  int early_release_inst;
  int early_release_reg;
  int restore_point;
  long return_stack[12];
  int rs_size;
  int rs_size_cp;
  int rs_size_sync_cp;
  int rs_start;
  int rs_start_cp;
  int rs_start_sync_cp;
  long return_stack_cp[12];
  long return_stack_sync_cp[12];
  unsigned long instrs, wpinstrs;
  long lasthazard[MAXREG];
  long lasthazard_cp[MAXREG];
  long lasthazard_sync_cp[MAXREG];
  long fetchcycle;
  int wait_for_fetch;
  struct context_t *mergethread;
  struct c_request *cache_entry;
  scoreboard_t score[SCORESIZE];
  int scoretop;
  int wrong_path;
  mem_addr rightpc;
  int writer[MAXREG];
  int writer_cp[MAXREG];
  int writer_sync_cp[MAXREG];
  int misfetch, misfetching, num_misfetches;
  int ghr;
  int ghr_cp;
  int ghr_sync_cp;
  int fthiscycle;
  int next_to_commit;

  mem_word *textseg, *dataseg, *stackseg;
  unsigned long textsize, datasize, stacksize;
  mem_addr pbreak, pstaddress;
  unsigned long stacktop;
  unsigned long stackbase;

  int id;

  /* If this context shares memory with another context(s),
     this integer specifies the original context number.
     If this context does not share memory, then this is
     equal to its context id. */
  int masterid;
  mem_addr srcmem, destmem;

  long sigstack;
  int sigonstack;
  long sigmask;

  FILE *infile, *outfile;
  FILE *argfile;

  Segment segment[65];

} context;

unsigned long available;

/* hazard types */
typedef enum { TopLevel=0, Fatal, Benign } EType;
enum { REG_V0, REG_A0=16, REG_A1, REG_A2, REG_A3, REG_A4, REG_A5, REG_ERR=7,
       REG_GP=29, REG_SP, REG_ZERO, REG_RA=26 };
enum { GENREGS = (1<<0), FPREGS = (1<<1) };
#ifndef RUNC
extern int prog_fds[];                  /* for fd translation */
#endif
/*
 * Find the size of the array
 */
#define asize(x)         (sizeof(x) / sizeof(x[0]))
#define ROUND(V, B) (((long) V + (B-1)) & ~(B-1))
/* Sign-extend a short to a long */
#define SEXT(X)        (((X) & 0x8000) ? ((X) | 0xffffffffffff0000) : (X))
#define SEXT21(X)        (((X) & 0x100000) ? ((X) | 0xffffffffffe00000) : (X))
#define SEXT32(X)        (((X) & 0x80000000) ? ((X) | 0xffffffff00000000) : (X))

#define SIGNEX(X)        (((X) & 0x8000) ? ((X) | 0xffffffffffff0000) : (X))
#define	STACK_ALIGN(x)	((unsigned)(x) &~ ((4*sizeof(int))-1))

/*
 * global definitions settable on the command line 
 */
extern int systrace, debug;
/* signals.c */
extern void setup_signal_stack(void);
extern void dosigreturn(context *);
extern void initsighandler(void);
extern void sighandle();
/* main.c */
/* mem.c -> mem.h */
extern void print_memstats(void);
/* ${SYSTEM}-syscall.c */
extern int dosyscall(context *);
/* print.c */
extern void print(int, unsigned long, unsigned int);
/* read-aout.c */
extern int read_symbols(char *);
extern int readaout(context *, char *, context *, int, int);
/* run.c */
extern int run();
extern void endofrun(void);
/* sched.c */
extern void initsched();
extern void schedstats(void);
extern void sched();
/* symbol.c */
extern void addsymbol(char *, int);
extern int findsymbol(char *);
extern int issymbol(char *);
extern void deletesymbols(void);
extern char *closest(int);
/* utils.c */
extern void error(EType, char *, ...);
extern void initfds(void);
extern void restorefds(void);
extern void *galloc(int);
extern void *grealloc(void *p, int size);
extern char *strsave(char *);
extern int strprefix(char *s, char *t, int c);

/* stuff in libc.a */
extern void bzero(char *, int);
extern int close(int), dup(int), dup2(int, int);
extern int syscall(int, ...);

#endif _MAIN_H_DEFINED

