/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <assert.h>
#include "main.h"
#include "mem.h"
#include <signal.h>
#include <setjmp.h>

extern struct sigvec sighandler[];
extern int sigaction_array[];
extern int prog_sigmask;
extern context *thecontexts[];

/*  I'm not even sure what works here and what doesn't.  The main
 *    goal now is that programs that don't rely too much on this
 *    should not break!
 */

extern void setupsignalstack(context *current,
			     int sig, int code, struct sigcontext *scp) {
    int i;
    mem_addr locsc;
    struct sigcontext *sc;
    
    locsc = (mem_addr) ((struct sigcontext *)STACK_ALIGN(current->R[30]) - 1);
    sc = (struct sigcontext *) mem_address(current, locsc);
    *sc = *scp;
    /*
     * we have to check if we are to put this on a signal stack or not 
     * scp->sc_onstack = ??;
     */
    sc->sc_mask = prog_sigmask;
    sc->sc_pc = current->pc;
    /* we save only that which is not saved by trampoline code */
    sc->sc_regs[REG_V0] = current->R[REG_V0];
    sc->sc_regs[REG_A0] = current->R[REG_A0];
    sc->sc_regs[REG_A1] = current->R[REG_A1];
    sc->sc_regs[REG_A2] = current->R[REG_A2];
    sc->sc_regs[REG_A3] = current->R[REG_A3];
    sc->sc_regs[REG_SP] = current->R[REG_SP];
    /*sc->sc_ownedfp = current->ownedfp;*/
	for(i=0; i < 32; ++i) /* FPU registers */
	    sc->sc_fpregs[i] = current->FWR[i];
	/*
	 * these are in the kernel code - I have no idea what to do with them
	 * sc->sc_fpc_eir = u.u_pcb.pcb_fpc_eir;
	 */
	sc->sc_fpcr = current->fpcr;
    current->R[30] = STACK_ALIGN(locsc);
    current->R[REG_A0] = sig;
    current->R[REG_A1] = code;
    current->R[REG_A2] = locsc;
    current->R[REG_A3] = (long) sighandler[current->R[REG_A0]].sv_handler;
}

extern void dosigreturn(context *current) {
    int i;
    struct sigcontext *sc;

    sc = (struct sigcontext *) mem_address(current, current->R[REG_A0]);
    prog_sigmask = sc->sc_mask;
    current->nextpc = sc->sc_pc;
    /*
     * This has to check for the branch delay bit (and someone has to set it)
     */
/*    current->nextpc = current->pc + 4; not sure why this is here */
    for(i=0; i < 32; ++i)
	current->R[i] = sc->sc_regs[i];
    for(i=0; i < 32; ++i)
	current->FWR[i] = sc->sc_fpregs[i];
    for(i=0; i < 32; ++i) /* FPU registers */
      current->FWR[i] = sc->sc_fpregs[i];
    current->fpcr = sc->sc_fpcr;
}

extern void sighandle(int sig, int code, struct sigcontext *scp) {
    /*
     * See if the process is blocking the signal
     * or if the CCR is set up to receive overflow/underflow/divbyzero
     * BUG BUG BUG I pulled these numbers out from nowhere - where are they
     * defined ? where is ccr[1] described ? Why is signal.h so bad ??
     */

  /*FIXME -- this is bound to not work.  Why not just delete it */

    context *current = thecontexts[0];
    if((prog_sigmask & (1 << current->R[REG_A0])) == 1)
	return;
#if 0
 ||
       ((sig == SIGFPE) && (((code == 0x20) && (!(current->CCR[1][31] & 0x200))) ||
			    ((code == 0x30) && (!(current->CCR[1][31] & 0x400))) ||
			    ((code == 0x40) && (!(current->CCR[1][31] & 0x800))))))
#endif
    if( sighandler[sig].sv_handler == 0) {
	/*
	 * Since the process is not trapping or blocking the signal,
	 * print a diagnostic and go interactive.
	 */
	switch(sig) {
	case SIGINT: error(Benign, "mipsi: Interrupt\n"); break;
	case SIGSEGV:error(Benign, "mipsi: Segmentation Fault\n"); break;
	case SIGBUS: error(Benign, "mipsi: Bus Error\n"); break;
	case SIGSYS: error(Benign, "mipsi: Bad arg to system call\n"); break;
	case SIGFPE: error(Benign, "mipsi: Floating Point exception\n"); break;
	default:     error(Benign, "mipsi: signal not caught\n"); break;
	}
	print(current->id, current->pc, 1);
	/*dumpregs(GENREGS | FPREGS);*/
        error(TopLevel, "Exception occurred at PC=0x%08x - no handler\n", current->pc);
    }
    setupsignalstack(current, sig, code, scp);
    if(sigtramp == 0)
	error(Fatal, "No sigvec call preceded the trapped signal ??\n");
    sigsetmask(sighandler[sig].sv_mask);
    /*FIXME this never set 
      longjmp(current->inrun, 1);
      */
}

int first_time = 1;
extern void dosigaction(context *current)
{
  int signum;
  
  if (first_time) {
    printf("ERROR: sigaction is not fully implemented.  If correct signal\n");
    printf("\tis invoked, the action will not actually be taken\n");
    first_time = 0;
  }
  signum = current->R[REG_A0];
  if (current->R[REG_A1]) {
    sigaction_array[signum] = current->R[REG_A1];
  }
  if (current->R[REG_A2]) {
    current->R[REG_A2] = sigaction_array[signum];
  }

  current->R[REG_V0] = 0;
  /* For some reason, __sigaction code expects A3 to have a 0 return value */
  current->R[REG_A3] = 0;
  
  /* Still need to add code so that on a signal, the 
     correct action is actually taken. */

  /* Still need to add support for returning the correct
     error messages (EFAULT, EINVAL) */
}
