/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Jack Lo.
Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
 */

#include "fetch.h"
#include "stash.h"
#include "main.h"
#include "cache.h"

#define SmtSTASH_ASSOC		1

#if defined(SmtSTASH_SMALL)
#define SmtSTASH_SIZE_VALUE	256	/* (2 ^ SmtSTASH_SIZE_EXPONENT) */
#define SmtSTASH_SIZE_EXPONENT	8
#elif defined(SmtSTASH_MEDIUM)
#define SmtSTASH_SIZE_VALUE	512	/* (2 ^ SmtSTASH_SIZE_EXPONENT) */
#define SmtSTASH_SIZE_EXPONENT	9
#else
#define SmtSTASH_SIZE_VALUE	1024	/* (2 ^ SmtSTASH_SIZE_EXPONENT) */
#define SmtSTASH_SIZE_EXPONENT	10
#endif

typedef struct StashItemStruct
{
	int valid[MAXTHREADS];
	unsigned int tags[MAXTHREADS];
	StashData data[MAXTHREADS][16];
} StashItem;

static StashItem items[SmtSTASH_ASSOC][SmtSTASH_SIZE_VALUE];

/*
** Stash statistics
*/
static unsigned long stash_full_count = 0;
static unsigned long stash_empty_count = 0;
static unsigned long stash_miss_count = 0;

StashData *stash;

void
init_stash()
{
	int line;
	int index;
	int thread;

	for(line = 0; line < SmtSTASH_ASSOC; line++)
	{
		for(index = 0; index < SmtSTASH_SIZE_VALUE; index++)
		{
			for(thread = 0; thread < MAXTHREADS; thread++)
			{
				items[line][index].tags[thread] = 0;
				items[line][index].valid[thread] = 0;
			}
		}
	}
}

/* FIXME -- very unintuitive, but works for now */

int get_stash_data(mem_addr addr, int thread, StashData **stashentry)
{
	StashItem *item;
	int index;
	int tag;
	int set;
	int line;
	int stash_full;
   
	index = (addr >> 2) & 15;
	tag = addr >> LINESIZEEXP;
	set = tag & (SmtSTASH_SIZE_VALUE - 1);
   
	for(line = 0; line < SmtSTASH_ASSOC; line++)
	{
		item = &items[line][set];
		if (item->tags[thread] == tag)
		{
		  *stashentry = &(item->data[thread][index]);
		  if (item->valid[thread] & (1 << index))
		    {
		      stash_full_count++;
		      return 1;
		    }
		  else
		    {
		      item->valid[thread] |= (1 << index);
		      stash_empty_count++;
		      return 0;
		    }
		}
	}

	/* FIXME  -- not associative yet */

	stash_miss_count++;
	item->valid[thread] = (1 << index);
	
	item->tags[thread] = tag;

	*stashentry = &item->data[thread][index];
	return(0);
}

void
print_stash_stats()
{
	int total = (stash_full_count + stash_empty_count + stash_miss_count);

	if(total != 0)
	{
		printf("STASH: size: %d entries, full %lu, ",
			(SmtSTASH_SIZE_VALUE * SmtSTASH_ASSOC * 8 * 8),
			stash_full_count);
		printf("empty %lu, misses %lu, full rate %.2f\n",
			stash_empty_count,
			stash_miss_count,
			((double) (100 * stash_full_count) / total));
	}
}

