/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdarg.h>
#define _BSD_SIGNALS  1
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include "stash.h"
#include "main.h"
#include "syscalls.h"
#include "mem.h"
#ifdef __alpha
#include <errno.h>
#include <sys/types.h>
#include <sys/mman.h>

#define handle_trap() asm(".set noat;call_pal 131;.set at;bis $0, $31, %0;":"=r"(retval)::"$1","$2","$3","$4","$5","$6","$7","$8","$16","$17","$18","$19","$20","$21","$22","$23","$24","$25","$28")

#ifdef 0
typedef struct {
        int     _cnt;
        unsigned char   *_ptr;
        unsigned char   *_base;
        int     _bufsiz;
        short   _flag;
        short   _file;
        int     _unused[2];
        void    *_lock;                 /* lock for thread safe library */
        unsigned char   *_bufendp;
} FILE;
#endif

#define _NIOBRW         8
extern FILE     _iob[_NIOBRW];

#define _IOEOF          0020
#define _IOERR          0040
#define stdin           (&_iob[0])
#define stdout          (&_iob[1])
#define stderr          (&_iob[2])
#else
#include <stdio.h>
#endif

extern int errno;
extern unsigned long data_start_address;

int prog_fds[MAXFDS]; /* anything that used prog_fds[] probably does not
			 work */
int prog_sigmask = 0;
mem_addr sigtramp = 0;
struct sigvec sighandler[NSIG];
long sigaction_array[NSIG] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

char kludgefilename1[12] = "texput.aux";
char kludgefilename2[12] = "temp.aux";

typedef long  byte8;
typedef int   byte4;
typedef short byte2;
struct alpha_stbuf {
    byte4   st_dev;
    byte4   st_ino;        
    byte4   st_mode;       
    byte2   st_nlink;      
    byte4   st_uid;        
    byte4   st_gid;        
    byte4   st_rdev;       
    byte8   st_size;        
    byte4   st_atime;       
    byte4   st_spare1;
    byte4   st_mtime;         
    byte4   st_spare2;
    byte4   st_ctime;         
    byte4   st_spare3;
    byte4   st_blksize;       
    byte4   st_blocks;         
    byte4   st_flags;         
    byte4   st_gen;           
};
struct ultrix_stbuf {
    byte2   st_dev;
    byte4   st_ino;        
    byte2   st_mode;       
    byte2   st_nlink;      
    byte2   st_uid;        
    byte2   st_gid;        
    byte2   st_rdev;       
    byte4   st_size;        
    byte4   st_atime;       
    byte4   st_spare1;
    byte4   st_mtime;         
    byte4   st_spare2;
    byte4   st_ctime;         
    byte4   st_spare3;
    byte4   st_blksize;       
    byte4   st_blocks;         
};


static char *mkstringarg(context *current, int argno, int regoff,
			 int reg, int comma) {
    static char buf[256];
    char *c, *bp, i;

    if(syscall_table[current->R[regoff]].args[argno] != NO_ARG) {
	if(syscall_table[current->R[regoff]].args[argno] == STR_ARG) {
	    if(comma) {buf[0]=',';buf[1]=' '; buf[2]='"'; buf[3]='\0';
		       bp = &buf[3];}
	    else      {buf[0]='"'; buf[1] = '\0'; bp = buf+1; }
	    c = (char *) mem_address(current, current->R[reg]);
	    for(i=0; (*c != '\0') && (i < 64); ++i)
		if(*c == '\n') { *bp++ = '\\'; *bp++ = 'n'; c++; }
		else  		 *bp++ = *c++;
	    *bp++ = '"';
	    *bp = '\0';
	}
	else 
	    sprintf(buf, "%s0x%x", (comma ? ", ": ""), current->R[reg]);
	return buf;
    }
    return "";
}

/* does the unix system call: returns negative on error */
static int unixsyscall(context *current) {
    mem_addr arg0, arg1, arg2, arg3, arg4, arg5, i;
    int _mprotect = 0;

    if(systrace) {
	traceout("SYSCALL %s(", syscall_table[current->R[REG_V0]].syscall_name);
	for(i=0; i < 4; ++i)
	    traceout(mkstringarg(current, i, REG_V0, REG_A0 + i, i != 0));
	traceout(") =  ");
    }

    arg0 = SYSCALL_ARG(REG_V0, 0, REG_A0);
    arg1 = SYSCALL_ARG(REG_V0, 1, REG_A1);
    arg2 = SYSCALL_ARG(REG_V0, 2, REG_A2);
    arg3 = SYSCALL_ARG(REG_V0, 3, REG_A3);
    arg4 = SYSCALL_ARG(REG_V0, 4, REG_A4);
    arg5 = SYSCALL_ARG(REG_V0, 5, REG_A5);
    if ((current->R[REG_V0] == 45) && strcmp(arg0,kludgefilename1) == 0 &&
	arg1 == 0x601) {
      arg0 = kludgefilename2;
    }
    if (current->R[REG_V0] == SYS_write && current->outfile != NULL
	&& arg0 == 1) {/*write to stdout*/
      arg0 = current->outfile->_file;
    }
    current->R[REG_V0] = syscall(current->R[REG_V0], arg0, arg1, arg2, arg3, arg4, arg5);
    traceout(" %d\n", current->R[REG_V0]);
    /* We have to see whether an error has occurred during
       the system call. If so, the libc wrap around must be
       notifified by setting register 7 to be less than zero
       and the return value should be set to errno. If not
       register 7 should be zero. r7 acts like the carry flag
       in the old days.
       */
    if((signed) current->R[REG_V0] < 0) {
      current->R[REG_A3] = -1;
/*    current->R[REG_A3] = errno; */
      return -1;
    } else {
/*
	current->R[REG_A3] = 0;
*/
	current->R[REG_A3] = 0;
	return current->R[REG_V0];
    }
}

/* does the reverse translation necessary for doing select */
static int reverse_fds(int fd) {
    int i;
    for(i=0; i < MAXFDS; ++i)
	if(prog_fds[i] == fd)
	    return i;
    error(Fatal, "Couldn't reverse translate fd %d\n",fd);
    return -1;
}

/*
 * decides which syscall it should do or simulate
 * returns non-zero to signal exit, zero for continue
 */
int dosyscall(context *current) {
    /* Use actual MIPS system calls after translating arguments
       from simulated memory to actual memory and to different
       file descriptors */
    struct ultrix_stbuf *ubuf;
    void *buf;
    int arg0;
    long retval;
    mem_addr addr_arg0;

#ifdef SYSVoffset
    current->R[REG_V0] -= SYSVoffset;
#endif

    /* Handle Mach kernel traps */
    switch(current->R[REG_V0]) {
    case -10:      /* task_self, trap -10, 0 args */
      traceout("SYSCALL task_self - ");
      asm("lda $0, -10($31)");
      handle_trap();
      break;
    case -11:      /* thread_reply, trap -11, 0 args */
      traceout("SYSCALL thread_reply - ");
      asm("lda $0, -11($31)");
      handle_trap();
      break;
    case -12:      /* task_notify, trap -12, 0 args */
      traceout("SYSCALL thread_notify - ");
      asm("lda $0, -12($31)");
      handle_trap();
      break;
    case -13:      /* thread_self, trap -13, 0 args */
      traceout("SYSCALL thread_self - ");
      asm("lda $0, -13($31)");
      handle_trap();
      break;
    case -20:      /* msg_send_trap, trap -20, 4 args */
      traceout("SYSCALL msg_send_trap - ");
      asm("lda $0, -20($31)");
      asm("ldq $16, %0;ldq $17, %1;ldq $18, %2;ldq $19, %3;"::"m" (current->R[REG_A0]),"m" (current->R[REG_A1]),"m" (current->R[REG_A2]),"m" (current->R[REG_A3]));
      handle_trap();
      break;
    case -21:      /* msg_receive_trap, trap -21, 5 args */
      traceout("SYSCALL msg_receive_trap - ");
      asm("lda $0, -20($31)");
      handle_trap();
      break;
    case -22:      /* msg_rpc_trap, trap -22, 6 args */
      traceout("SYSCALL msg_rpc_trap - ");
      addr_arg0 = mem_address_nullok(current, current->R[REG_A0]);
      asm("lda $0, -22($31)");
      asm("ldq $16, %0;ldq $17, %1;ldq $18, %2;ldq $19, %3;"::"m" (addr_arg0),"m" (current->R[REG_A1]),"m" (current->R[REG_A2]),"m" (current->R[REG_A3]));
      asm("ldq $20, %0;ldq $21, %1;"::"m" (current->R[REG_A4]),"m" (current->R[REG_A5]));
      handle_trap();
      break;
    case -30:      /* lw_wire, trap -30, 3 args */
      traceout("SYSCALL lw_wire - ");
      asm("lda $0, -30($31)");
      handle_trap();
      break;
    case -31:      /* lw_unwire, trap -31, 1 args */
      traceout("SYSCALL lw_unwire - ");
      asm("lda $0, -31($31)");
      handle_trap();
      break;
    case -52:      /* htg_unix_syscall, trap -52, 3 args */
      traceout("SYSCALL htg_unix_syscall - ");
      asm("lda $0, -52($31)");
      handle_trap();
      break;
    case -41:    /* init_process, trap -41, 0 args */
      traceout("SYSCALL init_process - ");
      asm("lda $0, -41($31)");
      handle_trap();
      break;
    case -43:    /* map_fd, trap -43, 5 args */
      traceout("SYSCALL map_fd - ");
      asm("lda $0, -43($31)");
      handle_trap();
      break;
    case -55:    /* host_self, trap -55, 1 args */
      traceout("SYSCALL host_self - ");
      asm("lda $0, -55($31)");
      handle_trap();
      break;
    case -56:    /* host_priv_self, trap -56, 1 args */
      traceout("SYSCALL host_priv_self - ");
      asm("lda $0, -56($31)");
      handle_trap();
      break;
    case -59:    /* swtch_pri, trap -59, 1 args */
      traceout("SYSCALL swtch_pri - ");
      asm("lda $0, -59($31)");
      handle_trap();
      break;
    case -60:    /* swtch, trap -60, 0 args */
      traceout("SYSCALL swtch - ");
      asm("lda $0, -60($31)");
      handle_trap();
      break;
    case -61:    /* thread_switch, trap -61, 3 args */
      traceout("SYSCALL thread_switch - ");
      asm("lda $0, -61($31)");
      handle_trap();
      break;
    }

    switch(current->R[REG_V0]) {
    case -10:    /* task_self, trap -10, 0 args */
    case -11:    /* thread_reply, trap -11, 0 args */
    case -12:    /* task_notify, trap -12, 0 args */
    case -13:    /* thread_self, trap -13, 0 args */
    case -20:    /* msg_send_trap, trap -20, 4 args */
    case -21:    /* msg_receive_trap, trap -21, 5 args */
    case -22:    /* msg_rpc_trap, trap -22, 6 args */
    case -30:    /* lw_wire, trap -30, 3 args */
    case -31:    /* lw_unwire, trap -31, 1 args */
    case -52:    /* htg_unix_syscall, trap -52, 3 args */
    case -41:    /* init_process, trap -41, 0 args */
    case -43:    /* map_fd, trap -43, 5 args */
    case -55:    /* host_self, trap -55, 1 args */
    case -56:    /* host_priv_self, trap -56, 1 args */
    case -59:    /* swtch_pri, trap -59, 1 args */
    case -60:    /* swtch, trap -60, 0 args */
    case -61:    /* thread_switch, trap -61, 3 args */
      if (retval < 0) {
	printf("Mach syscall %d returned error %d\n",
	       current->R[REG_V0], retval);
      }
      current->R[REG_V0] = retval;
      traceout(" %d\n", current->R[REG_V0]);
      return 0;
      break;
    }

    if ((current->R[REG_V0] == 0) &&
	(current->R[REG_A0] == 0x2000005))
      return 0;
    if (current->R[REG_V0] > MAX_SYSCALL)
	error(TopLevel, "Illegal system call: %d\n", current->R[REG_V0]);
    switch(syscall_table[current->R[REG_V0]].syscall_type) {
    case BAD_SYSCALL:
	error(TopLevel,"Unknown/Unimplemented system call: %s (%d) at %lx\n",
	      syscall_table[current->R[REG_V0]].syscall_name,current->R[REG_V0], current->pc);
	break;
    case UNIX_SYSCALL:
	unixsyscall(current);
	break;
    case SPC_SYSCALL:
	/* These special syscalls need to be simulated by SPIM */
	switch(current->R[REG_V0]) {
	case SYS_syscall:
	    traceout("SYSCALL syscall - ");
	    current->R[REG_V0] = current->R[REG_A0];
	    current->R[REG_A0] = current->R[REG_A1];
	    current->R[REG_A1] = current->R[REG_A2];
	    current->R[REG_A2] = current->R[REG_A3];
	    current->R[REG_A3] = rnr_mem_word(current, current->R[REG_SP]+16);
	    dosyscall(current);
	    break;
	case SYS_sysmips: {
	    /* 
	     * the table smipst goes from the sysmips arguments to syscall
	     * numbers
	     */
	    int callno, smipst[] = { MAX_SYSCALL+1, MAX_SYSCALL+1,
					 MAX_SYSCALL+1, MAX_SYSCALL+1,
					 MAX_SYSCALL+1, MAX_SYSCALL+1,
					 MAX_SYSCALL+1, MAX_SYSCALL+1,
					 MAX_SYSCALL+1, MAX_SYSCALL+1,
					 MAX_SYSCALL+1, MAX_SYSCALL+1,
					 SYS_getrusage, SYS_wait4, 
					 SYS_cacheflush, SYS_cachectl};
	    traceout("SYSCALL sysmips - ");
	    callno= current->R[REG_A0];
	    callno=((callno >0x100) ? smipst[callno - 0x100 + 10] : smipst[callno]);
	    current->R[REG_V0] = callno;
	    current->R[REG_A0] = current->R[REG_A1];
	    current->R[REG_A1] = current->R[REG_A2];
	    current->R[REG_A2] = current->R[REG_A3];
	    current->R[REG_A3] = rnr_mem_word(current, current->R[REG_SP]+16);
	    dosyscall(current);
	    break;
	}
	case SYS_exit: {
	    /* close all file descriptors */
	    int i;
	    traceout("SYSCALL exit() = ??\n");
/*	    for(i=0; i < MAXFDS; ++i)
		close(prog_fds[i]);*/
	    return (current->R[REG_A0] & 0xff) + 256;
	}
	case SYS_close: {
	    /* mark the file descriptor closed */
	    int ret;
	    if (current->R[REG_A0] < 3) {
	      current->R[REG_V0] = current->R[REG_A3] = 0;
	      break;
	    }
	    ret = unixsyscall(current);
	    if(ret >= 0)
		prog_fds[current->R[REG_A0]] = -1;
	    break;
	}      
	case SYS_read: {
	  traceout("SYSCALL read(%d, 0x%lx, 0x%lx) = ", current->R[REG_A0], 
		   current->R[REG_A1], current->R[REG_A2]);
	  if (current->R[REG_A0] == 0 && 
	      current->infile != NULL) {/*read from stdin*/
	    arg0 = current->infile->_file;
	  }
	  else arg0 = current->R[REG_A0];
	  current->R[REG_V0] = read(arg0, 
				    mem_address(current, current->R[REG_A1]),
				    current->R[REG_A2]);
	  if (current->R[REG_V0] != -1)
	    current->R[REG_A3] = 0;
	  else current->R[REG_A3] = -1;
	  traceout("%d\n", current->R[REG_V0]);
	  break;
	}      
	case SYS_socket: case SYS_open: case SYS_dup: {
	    /* update the fd translation table */
	    int ret = unixsyscall(current);
	    if(ret >= 0)
		prog_fds[ret] = ret;
	    break;
	}
	case SYS_dup2: {
	    int ret = unixsyscall(current);
	    if(ret >= 0)
		prog_fds[current->R[REG_A1]] = prog_fds[current->R[REG_A0]];
	    break;
	}
	case SYS_pipe: {
	    /* This isn't too useful unless we implement fork() or other
	       fd passing mechanisms */
	    int *fd;
	    if(unixsyscall(current) >= 0) {
		fd = (int *)mem_address(current, current->R[REG_A0]);
		prog_fds[fd[0]] = fd[0];
		prog_fds[fd[1]] = fd[1];
	    }
	    break;
	}
	case SYS_select: {
	    int fd;
	    /*
	     * we have to use this kludge to circumvent typechecking
	     * because the memory read macros take the lefthand side
	     * as an argument instead of simply returnign the value
	     * at the address
	     */
	    long int kludge;
	    fd_set a, b, c;
	    fd_set *readfds=&a, *writefds=&b, *exceptfds=&c;
	    struct timeval *timeout;
	    
	    FD_ZERO(readfds);
	    FD_ZERO(writefds);
	    FD_ZERO(exceptfds);
	    kludge = rnr_mem_word(current, current->R[REG_SP]+16);
	    if(kludge == NULL)
		timeout = NULL;
	    else
		timeout = (struct timeval *) mem_address(current, kludge);
	    if(current->R[REG_A1] == NULL)
		readfds = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, (fd_set *) mem_address(current, current->R[REG_A1])))
			FD_SET(prog_fds[fd], readfds);
	    if(current->R[REG_A2] == NULL)
		writefds = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, (fd_set *) mem_address(current, current->R[REG_A2])))
			FD_SET(prog_fds[fd], writefds);
	    if(current->R[REG_A3] == NULL)
		exceptfds = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, (fd_set *) mem_address(current, current->R[REG_A3])))
			FD_SET(prog_fds[fd], exceptfds);
	    traceout("SYSCALL select(%d, 0x%lx, 0x%lx, 0x%lx, 0x%lx) = ", current->R[REG_A0], current->R[REG_A1], current->R[REG_A2], current->R[REG_A3], timeout);
	    current->R[REG_V0] = select(current->R[REG_A0], readfds, writefds, exceptfds, timeout);
	    if(readfds == NULL)
		current->R[REG_A1] = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, readfds))
			FD_SET(reverse_fds(fd), 
			       (fd_set *) mem_address(current, current->R[REG_A1]));
	    if(writefds == NULL)
		current->R[REG_A2] = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, writefds))
			FD_SET(reverse_fds(fd),
			       (fd_set *) mem_address(current, current->R[REG_A2]));
	    if(exceptfds == NULL)
		current->R[REG_A3] = NULL;
	    else 
		for(fd=0; fd < current->R[REG_A0]; ++fd)
		    if(FD_ISSET(fd, exceptfds))
			FD_SET(reverse_fds(fd),
			       (fd_set *) mem_address(current, current->R[REG_A3]));
	    traceout("0x%lx\n", current->R[REG_V0]);
	    if((signed) current->R[REG_V0] < 0) {
		current->R[REG_A3] = -1;
/*		current->R[REG_A3] = errno; */
	    } else
		current->R[REG_A3] = 0;
	    break;
	}
	case SYS_sbrk:
	    traceout("SYSCALL sbrk(%d) = ", current->R[REG_A0]);
	    expand_data(current, current->R[REG_A0]);
	    current->R[REG_V0] = current->pbreak;
	    current->pbreak += current->R[REG_A0];
	    current->R[REG_A3] = 0;
	    traceout("0x%lx\n", current->R[REG_V0]);
	    break;
	case SYS_obreak: {
	    unsigned long datatop = data_start_address + current->datasize;
	    traceout("SYSCALL brk(0x%lx) = ", current->R[REG_A0]);
	    /* Round up to PAGESIZE boundary */
	    if((current->R[REG_A0] - datatop) > 0)
		expand_data(current,
			    ROUND(current->R[REG_A0], PAGESIZE) - datatop);
	    /*
	     * if we get here, expand data must have been successful
	     */
	    current->R[REG_V0] = current->pbreak;
	    current->pbreak = ROUND(current->R[REG_A0], PAGESIZE);
	    current->R[REG_A3] = 0;
	    traceout("0x%lx\n", current->R[REG_V0]);
	    break;
	}
	case SYS_setitimer: {
	    struct itimerval it, oit;

	    traceout("SYSCALL setitimer(%d, 0x%lx, 0x%lx) = ", current->R[REG_A0],
		     current->R[REG_A1], current->R[REG_A2]);
	    it = *((struct itimerval *) mem_address(current, current->R[REG_A1]));
	    current->R[REG_V0] = setitimer(current->R[REG_A0], &it, mem_address_nullok(current, current->R[REG_A2]));
	    current->R[REG_A3] = ((signed)current->R[REG_V0] < 0) ? -1 : 0;
	    traceout("0x%lx\n", current->R[REG_V0]);
	    break;
	}
	case SYS_setrlimit:
	    traceout("SYSCALL setrlimit(...) = 0\n", current->R[REG_A0]);
	    /* as is evident, we fool it to think that we set some limits */
	    current->R[REG_V0] = current->R[REG_A3] = 0; 
	    break;
	case SYS_sigstack:
	    traceout("SYSCALL sigstack(0x%lx, 0x%lx) = 0\n", current->R[REG_A0], current->R[REG_A1]);
	    current->R[REG_V0] = current->R[REG_A3] = 0;
	    if (current->R[REG_A1] != 0) {
	      set_mem_quad(current, current->R[REG_A1],current->sigstack);
	      set_mem_word(current, current->R[REG_A1]+4,current->sigonstack);
	    }
	    if (current->R[REG_A0] != 0) {
	      current->sigstack = read_mem_quad(current, current->R[REG_A0]);
	      current->sigonstack = read_mem_word(current, current->R[REG_A0]+8);
	    }
	    break;
#ifdef 0
	case SYS_sigvec: { 
	    struct sigvec hnd = { sighandle, 0, 0};
	    int x;
	    
	    traceout("SYSCALL sigvec(%d, 0x%lx, 0x%lx) = ", current->R[REG_A0], current->R[REG_A1], current->R[REG_A2]);
	    if(current->R[REG_A2] != 0) {
		snr_mem_word(current, current->R[REG_A2], sighandler[current->R[REG_A0]].sv_handler);
		snr_mem_word(current, current->R[REG_A2]+sizeof(int*),sighandler[current->R[REG_A0]].sv_mask);
		snr_mem_word(current, current->R[REG_A2]+sizeof(int *)+sizeof(sigset_t), sighandler[current->R[REG_A0]].sv_flags);
	    }
	    if((current->R[REG_A1] == 0) || (current->R[REG_A1] == 1)) /* SIG_DFL SIG_IGN */
		current->R[REG_V0] = sigvec(current->R[REG_A0], current->R[REG_A1],NULL);
	    else {
		x = rnr_mem_word(current, current->R[REG_A1]);
		sighandler[current->R[REG_A0]].sv_handler = (void(*)()) x;
		x = rnr_mem_word(current, current->R[REG_A1] + sizeof (int *));
		sighandler[current->R[REG_A0]].sv_mask = x;
		x = rnr_mem_word(current, current->R[REG_A1] + sizeof(int *) + sizeof(sigset_t));
		sighandler[current->R[REG_A0]].sv_flags = x;
		current->R[REG_V0] = sigvec(current->R[REG_A0], &hnd, NULL);
	    }
	    sigtramp = current->R[REG_V0];
	    current->R[REG_A3] = ((signed)current->R[REG_V0] < 0) ? -1 : 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	}
#endif
	case SYS_sigreturn:
	    traceout("SYSCALL sigreturn(0x%lx) = ",current->R[REG_A0]);
	    dosigreturn(current); /* FIXME -- unlikely to be right */
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	case SYS_sigaction:
	    traceout("SYSCALL sigaction(%d,0x%lx,0x%lx) = ",current->R[REG_A0],
		     current->R[REG_A1],current->R[REG_A2]);
	    dosigaction(current);
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	case SYS_sigprocmask:
	    traceout("SYSCALL sigprocmask(%d, %lx, %lx) = ", current->R[REG_A0],
		     current->R[REG_A1],current->R[REG_A2]);
	    if (current->R[REG_A2] > 0x10000000)
	      set_mem_quad(current, current->R[REG_A2],current->sigmask);
	    if (current->R[REG_A1] != 0) {
	      switch(current->R[REG_A0]) {
	      case SIG_BLOCK:
		current->sigmask |= current->R[REG_A1];
		break;
	      case SIG_UNBLOCK:
		current->sigmask &= current->R[REG_A1]; /* I think */
		break;
	      case SIG_SETMASK:
		current->sigmask = current->R[REG_A1]; /* I think */
		break;
	      default:
		error(TopLevel,"illegal how value to sigprocmask\n");
		break;
	      }
	    }
	    current->R[REG_V0] = 0;
	    current->R[REG_A3] = 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
#ifdef 0
	case SYS_sigsetmask:
	    traceout("SYSCALL sigsetmask(0x%lx) = ", current->R[REG_A0]);
	    current->R[REG_V0] = prog_sigmask;
	    prog_sigmask = current->R[REG_A0];
	    traceout("%d\n", current->R[REG_V0]);
	    current->R[REG_A3] = 0;
	    break;
	case SYS_sigblock:
	    traceout("SYSCALL sigsetmask(0x%lx) = ", current->R[REG_A0]);
	    current->R[REG_V0] = prog_sigmask;
	    prog_sigmask |= current->R[REG_A0];
	    traceout("%d\n", current->R[REG_V0]);
	    current->R[REG_A3] = 0;
	    break;
#endif
	case SYS_cacheflush:
	    traceout("SYSCALL cacheflush(0x%lx, 0x%lx, 0x%lx) = ",
		     current->R[REG_A0],current->R[REG_A1],current->R[REG_A2]);
	    current->R[REG_V0] = 
		cache_flush((void*)mem_address(current, current->R[REG_A0]),current->R[REG_A1],current->R[REG_A2]);
	    current->R[REG_A3] = 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	case SYS_cachectl:
	    traceout("SYSCALL cachectl(0x%lx, 0x%lx, 0x%lx) = ", 
		     current->R[REG_A0],current->R[REG_A1],current->R[REG_A2]);
	    current->R[REG_V0] = 
		cache_ctl((void*)mem_address(current, current->R[REG_A0]),current->R[REG_A1],current->R[REG_A2]);
	    current->R[REG_A3] = 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	    /* the following are here just because the system call
	     * numbers have changed on the alpha
	     */
	case SYS_fstat:
	    traceout("SYSCALL fstat(%d, 0x%lx) = ", current->R[REG_A0], mem_address(current, current->R[REG_A1]));
	    goto dostat;
	case SYS_lstat:
	case SYS_stat:
	    traceout("SYSCALL %cstat(%s, 0x%lx) = ", 
		     current->R[REG_V0] == SYS_lstat ? 'l' : ' ',
		     mem_address(current, current->R[REG_A0]),
		     mem_address(current, current->R[REG_A1]));
	dostat:

	    buf = mem_address(current, current->R[REG_A1]); 
	    if(current->R[REG_V0] == SYS_fstat)
		current->R[REG_V0] = fstat(current->R[REG_A0], buf);
	    else if(current->R[REG_V0] == SYS_lstat)
		current->R[REG_V0] = lstat(mem_address(current, current->R[REG_A0]), buf);
	    else
		current->R[REG_V0] = stat(mem_address(current, current->R[REG_A0]), buf);
	    current->R[REG_A3] = ((signed)current->R[REG_V0] < 0) ? -1 : 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	case SYS_getpagesize:
	    traceout("SYSCALL getpagesize() = %d\n", PAGESIZE);
	    current->R[REG_V0] = PAGESIZE;
	    current->R[REG_A3] = 0;
	    break;
	case SYS_ioctl:
	    traceout("SYSCALL ioctl(%d, %lx, %lx) = ", current->R[REG_A0], current->R[REG_A1],
		     mem_address(current, current->R[REG_A2]));
#ifdef __alpha
	    if(current->R[REG_A1] == 0x40247455) /* TCGETP on ultrix */
		current->R[REG_A1] = 0x40067408; /* TIOCGETP on OSF1.3 */
#endif
	    /*kludge*/
	    if (current->R[REG_A1] == 0x40067408) {
	      current->R[REG_V0] = ioctl(0, current->R[REG_A1], mem_address(current, current->R[REG_A2]));
	    }
	    else if (current->R[REG_A0] == 0 && current->infile != NULL) {
	      current->R[REG_V0] = ioctl(current->infile->_file, current->R[REG_A1], mem_address(current, current->R[REG_A2]));
	    }
	    else if (current->R[REG_A0] == 1 && current->outfile != NULL) {
	      current->R[REG_V0] = ioctl(current->outfile->_file, current->R[REG_A1], mem_address(current, current->R[REG_A2]));
	    }
	    else
	      current->R[REG_V0] = ioctl(current->R[REG_A0], current->R[REG_A1], mem_address(current, current->R[REG_A2]));
	    if (current->R[REG_V0]) { /* MAJOR KLUDGE */
	      current->R[REG_V0] = 0;
	      current->R[REG_A3] = 1;
	    }
	    else current->R[REG_A3] = 0;
	    traceout("%d\n", current->R[REG_V0]);
	    break;
	  case SYS_uswitch:
	    traceout("SYSCALL uswitch(%d, %d) = ", current->R[REG_A0], current->R[REG_A1]);
	    current->R[REG_V0] = current->R[REG_A1]; /*not really implemented, obviously*/
	    traceout("%d\n", current->R[REG_V0]);
	    /*makezeroseg();*/
	    break;
	default:
	    error(Fatal, "Unimplemented special system call: %d\n",current->R[REG_V0]);
	    break;
	}
	break;
    default:
	error(Fatal, "Unknown type of syscall: %d\n", current->R[REG_V0]);
	break;
    }
    return 0;
}
